/*
 * Copyright (c) 2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

// CLASS HEADER
#include <dali-toolkit/internal/text/text-scroller.h>

// EXTERNAL INCLUDES
#include <dali/public-api/common/stage.h>
#include <dali/integration-api/debug.h>

// INTERNAL INCLUDES
#include <dali-toolkit/internal/text/text-scroller-interface.h>

namespace Dali
{

namespace Toolkit
{

namespace
{

#if defined ( DEBUG_ENABLED )
  Debug::Filter* gLogFilter = Debug::Filter::New(Debug::NoLogging, true, "LOG_TEXT_SCROLLING");
#endif

const int MINIMUM_SCROLL_SPEED = 1; // Speed should be set by Property system.

const char* VERTEX_SHADER_SCROLL = DALI_COMPOSE_SHADER(
  attribute mediump vec2 aPosition;\n
  varying highp vec2 vTexCoord;\n
  varying highp float vRatio;\n
  uniform mediump mat4 uMvpMatrix;\n
  uniform mediump vec3 uSize;\n
  uniform mediump float uDelta;\n
  uniform mediump vec2 uTextureSize;\n
  uniform mediump float uGap;\n
  uniform mediump float uHorizontalAlign;\n
  uniform mediump float uVerticalAlign;\n
  \n
  void main()\n
  {\n
    {\n
      float smallTextPadding = max( uSize.x - uTextureSize.x, 0. );\n
      float gap = max( uGap, smallTextPadding );\n
      float delta = floor ( uDelta ) + 0.5;\n
      vTexCoord.x = ( delta + uHorizontalAlign * ( uTextureSize.x - uSize.x ) + floor( aPosition.x * uSize.x ) + 0.5 - gap * 0.5 ) / (uTextureSize.x + gap) + 0.5;\n
      vTexCoord.y = ( uVerticalAlign * ( uTextureSize.y - uSize.y ) + floor( aPosition.y * uSize.y ) + 0.5 ) / ( uTextureSize.y ) + 0.5;\n
      vRatio = uTextureSize.x / ( uTextureSize.x + gap );\n
      gl_Position = uMvpMatrix * vec4( floor( aPosition * uSize.xy ), 0.0, 1.0 );\n
    }\n
  }\n
);

const char* FRAGMENT_SHADER = DALI_COMPOSE_SHADER(
  varying highp vec2 vTexCoord;\n
  varying highp float vRatio;\n
  uniform sampler2D sTexture;\n
  \n
  void main()\n
  {\n
    highp vec2 texCoord;\n
    texCoord.y = vTexCoord.y;\n
    texCoord.x = fract( vTexCoord.x ) / vRatio;\n
    if ( texCoord.x > 1.0 || texCoord.y > 1.0 )\n
      discard;\n
    \n
    gl_FragColor = texture2D( sTexture, texCoord );\n
  }\n
);

/**
 * @brief How the text should be aligned horizontally when scrolling the text.
 *
 * -0.5f aligns the text to the left, 0.0f aligns the text to the center, 0.5f aligns the text to the right.
 * The final alignment depends on two factors:
 *   1) The alignment value of the text label (Use Text::Layout::HorizontalAlignment enumerations).
 *   2) The text direction, i.e. whether it's LTR or RTL (0 = LTR, 1 = RTL).
 */
const float HORIZONTAL_ALIGNMENT_TABLE[ Text::Layout::HORIZONTAL_ALIGN_COUNT ][ 2 ] =
{
  // HORIZONTAL_ALIGN_BEGIN
  {
    -0.5f, // LTR
    0.5f   // RTL
  },

  // HORIZONTAL_ALIGN_CENTER
  {
    0.0f,  // LTR
    0.0f   // RTL
  },

  // HORIZONTAL_ALIGN_END
  {
    0.5f,  // LTR
    -0.5f  // RTL
  }
};

/**
 * @brief How the text should be aligned vertically when scrolling the text.
 *
 * -0.5f aligns the text to the top, 0.0f aligns the text to the center, 0.5f aligns the text to the bottom.
 * The alignment depends on the alignment value of the text label (Use Text::Layout::VerticalAlignment enumerations).
 */
const float VERTICAL_ALIGNMENT_TABLE[ Text::Layout::VERTICAL_ALIGN_COUNT ] =
{
  -0.5f, // VERTICAL_ALIGN_TOP
  0.0f,  // VERTICAL_ALIGN_CENTER
  0.5f   // VERTICAL_ALIGN_BOTTOM
};

} // namespace

namespace Text
{

TextScrollerPtr TextScroller::New( ScrollerInterface& scrollerInterface )
{
  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller::New\n" );

  TextScrollerPtr textScroller( new TextScroller( scrollerInterface) );
  return textScroller;
}

void TextScroller::SetGap( int gap )
{
  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller::SetGap gap[%d]\n", gap );
  mWrapGap = static_cast<float>(gap);
}

int TextScroller::GetGap() const
{
  return static_cast<int>(mWrapGap);
}

void TextScroller::SetSpeed( int scrollSpeed )
{
  mScrollSpeed = std::max( MINIMUM_SCROLL_SPEED, scrollSpeed );
}

int TextScroller::GetSpeed() const
{
  return mScrollSpeed;
}

void TextScroller::SetLoopCount( int loopCount )
{
  if ( loopCount >= 0 )
  {
    mLoopCount = loopCount;
  }

  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller::SetLoopCount [%d] Status[%s]\n", mLoopCount, (loopCount)?"looping":"stop" );
}

int TextScroller::GetLoopCount() const
{
  return mLoopCount;
}

void TextScroller::SetLoopDelay( float delay )
{
  mLoopDelay = delay;
}

float TextScroller::GetLoopDelay() const
{
  return mLoopDelay;
}

void TextScroller::SetStopMode( DevelTextLabel::AutoScrollStopMode::Type stopMode )
{
  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller::SetAutoScrollStopMode [%s]\n",(stopMode == DevelTextLabel::AutoScrollStopMode::IMMEDIATE)?"IMMEDIATE":"FINISH_LOOP" );
  mStopMode = stopMode;
}

void TextScroller::StopScrolling()
{
  if ( mScrollAnimation && mScrollAnimation.GetState() == Animation::PLAYING )
  {
    switch( mStopMode )
    {
      case DevelTextLabel::AutoScrollStopMode::IMMEDIATE:
      {
        mScrollAnimation.Stop();
        mScrollerInterface.ScrollingFinished();
        break;
      }
      case DevelTextLabel::AutoScrollStopMode::FINISH_LOOP:
      {
        mScrollAnimation.SetLoopCount( 1 ); // As animation already playing this allows the current animation to finish instead of trying to stop mid-way
        break;
      }
      default:
      {
        DALI_LOG_INFO( gLogFilter, Debug::Verbose, "Undifined AutoScrollStopMode\n" );
      }
    }
  }
}

DevelTextLabel::AutoScrollStopMode::Type TextScroller::GetStopMode() const
{
  return mStopMode;
}

TextScroller::TextScroller( ScrollerInterface& scrollerInterface )
: mScrollerInterface( scrollerInterface ),
  mScrollDeltaIndex( Property::INVALID_INDEX ),
  mScrollSpeed( MINIMUM_SCROLL_SPEED ),
  mLoopCount( 1 ),
  mLoopDelay( 0.0f ),
  mWrapGap( 0.0f ),
  mStopMode( DevelTextLabel::AutoScrollStopMode::FINISH_LOOP )
{
  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller Default Constructor\n" );
}

TextScroller::~TextScroller()
{
}

void TextScroller::SetParameters( Actor scrollingTextActor, Renderer renderer, TextureSet textureSet, const Size& controlSize, const Size& textNaturalSize, CharacterDirection direction, Layout::HorizontalAlignment horizontalAlignment, Layout::VerticalAlignment verticalAlignment )
{
  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller::SetParameters controlSize[%f,%f] offscreenSize[%f,%f] direction[%d]\n",
                 controlSize.x, controlSize.y, textNaturalSize.x, textNaturalSize.y, direction );

  mRenderer = renderer;

  float animationProgress = 0.0f;
  int   remainedLoop = mLoopCount;
  if ( mScrollAnimation )
  {
    if( mScrollAnimation.GetState() == Animation::PLAYING )
    {
      animationProgress = mScrollAnimation.GetCurrentProgress();

      if( mLoopCount > 0 ) // If not a ininity loop, then calculate remained loop
      {
        remainedLoop = mLoopCount - ( mScrollAnimation.GetCurrentLoop() );
        remainedLoop = ( remainedLoop <= 0 ? 1 : remainedLoop );
      }
    }
    mScrollAnimation.Clear();

    // Reset to the original shader and texture before scrolling
    mRenderer.SetShader(mShader);
    mRenderer.SetTextures( mTextureSet );
  }

  mShader = mRenderer.GetShader();
  mTextureSet = mRenderer.GetTextures();

  // Set the shader and texture for scrolling
  Shader shader = Shader::New( VERTEX_SHADER_SCROLL, FRAGMENT_SHADER, Shader::Hint::NONE );
  mRenderer.SetShader( shader );
  mRenderer.SetTextures( textureSet );

  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller::SetParameters mWrapGap[%f]\n", mWrapGap );

  const float horizontalAlign = HORIZONTAL_ALIGNMENT_TABLE[ horizontalAlignment ][ direction ];
  const float verticalAlign = VERTICAL_ALIGNMENT_TABLE[ verticalAlignment ];
  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller::SetParameters horizontalAlign[%f], verticalAlign[%f]\n", horizontalAlign, verticalAlign );

  scrollingTextActor.RegisterProperty( "uTextureSize", textNaturalSize );
  scrollingTextActor.RegisterProperty( "uHorizontalAlign", horizontalAlign );
  scrollingTextActor.RegisterProperty( "uVerticalAlign", verticalAlign );
  scrollingTextActor.RegisterProperty( "uGap", mWrapGap );
  mScrollDeltaIndex = scrollingTextActor.RegisterProperty( "uDelta", 0.0f );

  float scrollAmount = std::max( textNaturalSize.width + mWrapGap, controlSize.width );
  float scrollDuration =  scrollAmount / mScrollSpeed;

  if ( direction  )
  {
     scrollAmount = -scrollAmount; // reverse direction of scrolling
  }

  StartScrolling( scrollingTextActor, scrollAmount, scrollDuration, remainedLoop );
  mScrollAnimation.SetCurrentProgress(animationProgress);
}

void TextScroller::AutoScrollAnimationFinished( Dali::Animation& animation )
{
  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller::AutoScrollAnimationFinished\n" );
  mScrollerInterface.ScrollingFinished();

  // Revert to the original shader and texture after scrolling
  mRenderer.SetShader(mShader);
  if ( mTextureSet )
  {
    mRenderer.SetTextures( mTextureSet );
  }
}

void TextScroller::StartScrolling( Actor scrollingTextActor, float scrollAmount, float scrollDuration, int loopCount )
{
  DALI_LOG_INFO( gLogFilter, Debug::Verbose, "TextScroller::StartScrolling scrollAmount[%f] scrollDuration[%f], loop[%d] speed[%d]\n", scrollAmount, scrollDuration, loopCount, mScrollSpeed );

  mScrollAnimation = Animation::New( scrollDuration );
  mScrollAnimation.AnimateTo( Property( scrollingTextActor, mScrollDeltaIndex ), scrollAmount, TimePeriod( mLoopDelay, scrollDuration ) );
  mScrollAnimation.SetEndAction( Animation::Discard );
  mScrollAnimation.SetLoopCount( loopCount );
  mScrollAnimation.FinishedSignal().Connect( this, &TextScroller::AutoScrollAnimationFinished );
  mScrollAnimation.Play();
}

} // namespace Text

} // namespace Toolkit

} // namespace Dali
